@@ -1,23 +1,30 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import csv
-import sys
-from io import BytesIO, StringIO
+from io import StringIO
 
 from sql import Table
+from sql.aggregate import Sum
+from sql.conditionals import Coalesce
 
+from trytond.config import config
 from trytond.pool import PoolMeta, Pool
+from trytond.pyson import Eval
 from trytond.transaction import Transaction
 from trytond.wizard import Wizard, StateView, StateTransition, Button
 from trytond.model import ModelView, ModelStorage, fields
 
+OPENING_CODE = config.get('account_fr', 'fec_opening_code', default="OUV")
+OPENING_NAME = config.get(
+    'account_fr', 'fec_opening_name', default="Balance Initiale")
+OPENING_NUMBER = config.get('account_fr', 'fec_opening_number', default="0")
+
 
 __all__ = ['TaxTemplate', 'TaxRuleTemplate',
     'AccountFrFEC', 'AccountFrFECStart', 'AccountFrFECResult']
 
 
-class TaxTemplate:
-    __metaclass__ = PoolMeta
+class TaxTemplate(metaclass=PoolMeta):
     __name__ = 'account.tax.template'
 
     @classmethod
@@ -53,8 +60,7 @@ class TaxTemplate:
         super(TaxTemplate, cls).__register__(module_name)
 
 
-class TaxRuleTemplate:
-    __metaclass__ = PoolMeta
+class TaxRuleTemplate(metaclass=PoolMeta):
     __name__ = 'account.tax.rule.template'
 
     @classmethod
@@ -94,10 +100,7 @@ class AccountFrFEC(Wizard):
             ])
 
     def transition_generate(self):
-        if sys.version_info < (3,):
-            fec = BytesIO()
-        else:
-            fec = StringIO()
+        fec = StringIO()
         writer = self.get_writer(fec)
         writer.writerow(self.get_header())
         format_date = self.get_format_date()
@@ -106,17 +109,13 @@ class AccountFrFEC(Wizard):
         def replace_delimiter(c):
             delimiter = writer.dialect.delimiter
             return (c or '').replace(delimiter, ' ')
-        if sys.version_info < (3,):
-            def convert(c):
-                return replace_delimiter(c).encode('utf-8')
-        else:
-            convert = replace_delimiter
+        convert = replace_delimiter
 
         for row in self.get_start_balance():
-            writer.writerow(map(convert, row))
+            writer.writerow(list(map(convert, row)))
         for line in self.get_lines():
             row = self.get_row(line, format_date, format_number)
-            writer.writerow(map(convert, row))
+            writer.writerow(list(map(convert, row)))
         value = fec.getvalue()
         if not isinstance(value, bytes):
             value = value.encode('utf-8')
@@ -127,10 +126,13 @@ class AccountFrFEC(Wizard):
         file_ = self.result.file
         self.result.file = None  # No need to store it in session
         format_date = self.get_format_date()
-        filename = '%sFEC%s.csv' % (
-            self.start.fiscalyear.company.party.siren or '',
-            format_date(self.start.fiscalyear.end_date),
-            )
+        if self.start.fiscalyear.state == 'locked':
+            filename = '%sFEC%s.csv' % (
+                self.start.fiscalyear.company.party.siren or '',
+                format_date(self.start.fiscalyear.end_date),
+                )
+        else:
+            filename = None
         return {
             'file': file_,
             'filename': filename,
@@ -144,7 +146,8 @@ class AccountFrFEC(Wizard):
     def get_format_date(self):
         pool = Pool()
         Lang = pool.get('ir.lang')
-        return lambda value: Lang.strftime(value, 'fr_FR', '%Y%m%d')
+        fr = Lang(code='fr_FR')
+        return lambda value: fr.strftime(value, '%Y%m%d')
 
     def get_format_number(self):
         pool = Pool()
@@ -154,12 +157,12 @@ class AccountFrFEC(Wizard):
             thousands_sep='',
             grouping='[]',
             )
-        return lambda value: Lang.format(fr, '%.2f', value)
+        return lambda value: fr.format('%.2f', value)
 
     def get_header(self):
         return [
-            'JournalCode ',
-            'JournalLib ',
+            'JournalCode',
+            'JournalLib',
             'EcritureNum',
             'EcritureDate',
             'CompteNum',
@@ -181,33 +184,66 @@ class AccountFrFEC(Wizard):
     def get_start_balance(self):
         pool = Pool()
         Account = pool.get('account.account')
+        Move = pool.get('account.move')
+        Line = pool.get('account.move.line')
+        FiscalYear = pool.get('account.fiscalyear')
+        Period = pool.get('account.period')
+        Party = pool.get('party.party')
+        account = Account.__table__()
+        move = Move.__table__()
+        line = Line.__table__()
+        period = Period.__table__()
+        party = Party.__table__()
+        cursor = Transaction().connection.cursor()
         format_date = self.get_format_date()
         format_number = self.get_format_number()
 
-        with Transaction().set_context(
-                periods=[-1],
-                fiscalyear=self.start.fiscalyear.id,
-                posted=True,
-                cumulate=True):
-            accounts = Account.search([])
+        company = self.start.fiscalyear.company
+        fiscalyears = FiscalYear.search([
+                ('end_date', '<', self.start.fiscalyear.start_date),
+                ('company', '=', company.id),
+                ])
+        fiscalyear_ids = list(map(int, fiscalyears))
+        if not fiscalyear_ids:
+            return
 
-        for account in accounts:
-            if account.credit == account.debit:
-                continue
-            if account.debit > account.credit:
-                debit, credit = account.debit - account.credit, 0
-            else:
-                debit, credit = 0, account.debit - account.credit
-            yield [
-                self.start.deferral_journal.code
-                or self.start.deferral_journal.name,
-                self.start.deferral_journal.name,
-                self.start.deferral_post_number,
-                format_date(self.start.fiscalyear.start_date),
+        query = (account
+            .join(line, condition=line.account == account.id)
+            .join(move, condition=line.move == move.id)
+            .join(period, condition=move.period == period.id)
+            .join(party, 'LEFT', condition=line.party == party.id)
+            .select(
+                account.id,
                 account.code,
                 account.name,
-                '',
-                '',
+                party.code,
+                party.name,
+                Sum(Coalesce(line.debit, 0) - Coalesce(line.credit, 0)),
+                where=(account.company == company.id)
+                & (move.state == 'posted')
+                & (line.state != 'draft')
+                & period.fiscalyear.in_(fiscalyear_ids),
+                group_by=[
+                    account.id, account.code, account.name,
+                    party.code, party.name]))
+        cursor.execute(*query)
+        for row in cursor:
+            _, code, name, party_code, party_name, balance = row
+            if not balance:
+                continue
+            if balance > 0:
+                debit, credit = balance, 0
+            else:
+                debit, credit = 0, -balance
+            yield [
+                OPENING_CODE,
+                OPENING_NAME,
+                OPENING_NUMBER,
+                format_date(self.start.fiscalyear.start_date),
+                code,
+                name,
+                party_code or '',
+                party_name or '',
                 '-',
                 format_date(self.start.fiscalyear.start_date),
                 '-',
@@ -224,10 +260,15 @@ class AccountFrFEC(Wizard):
         pool = Pool()
         Line = pool.get('account.move.line')
 
-        return Line.search([
-                ('move.period.fiscalyear', '=', self.start.fiscalyear.id),
-                ('move.state', '=', 'posted'),
-                ],
+        domain = [
+            ('move.period.fiscalyear', '=', self.start.fiscalyear.id),
+            ('move.state', '=', 'posted'),
+            ['OR', ('debit', '!=', 0), ('credit', '!=', 0)],
+            ]
+        if self.start.deferral_period:
+            domain.append(('move.period', '!=', self.start.deferral_period.id))
+        return Line.search(
+            domain,
             order=[
                 ('move.post_number', 'ASC'),
                 ])
@@ -267,13 +308,6 @@ class AccountFrFEC(Wizard):
             ]
 
     def get_reference(self, line):
-        pool = Pool()
-        try:
-            Invoice = pool.get('account.invoice')
-        except KeyError:
-            Invoice = None
-        if Invoice and isinstance(line.move.origin, Invoice):
-            return line.move.origin.number
         if isinstance(line.move.origin, ModelStorage):
             return line.move.origin.rec_name
         return line.move.origin
@@ -293,27 +327,25 @@ class AccountFrFECStart(ModelView):
     'Generate FEC'
     __name__ = 'account.fr.fec.start'
 
-    fiscalyear = fields.Many2One('account.fiscalyear', 'Fiscal Year',
-        required=True, domain=[
-            ('state', '=', 'close'),
-            ])
+    fiscalyear = fields.Many2One(
+        'account.fiscalyear', 'Fiscal Year', required=True)
     type = fields.Selection([
             ('is-bic', 'IS-BIC'),
             ], 'Type', required=True)
-    deferral_journal = fields.Many2One('account.journal',
-        'Deferral Journal', required=True,
-        help='Journal used for pseudo deferral move')
-    deferral_post_number = fields.Char('Deferral Number', required=True,
-        help='Post number used for pseudo deferral move')
+    deferral_period = fields.Many2One(
+        'account.period', "Deferral Period", required=True,
+        domain=[
+            ('fiscalyear', '=', Eval('fiscalyear')),
+            ('type', '=', 'adjustment'),
+            ],
+        depends=['fiscalyear'],
+        help="The period to exclude which contains "
+        "the moves to balance non-deferral account")
 
     @classmethod
     def default_type(cls):
         return 'is-bic'
 
-    @classmethod
-    def default_deferral_post_number(cls):
-        return '0'
-
 
 class AccountFrFECResult(ModelView):
     'Generate FEC'
